;;; This file is part of guile-with-graphics, a graphic library for Guile

;;; Copyright (C) 2009  Luca Saiu

;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;;; Draw a *non-filled* ractangle. The interface is compatible with
;;; draw-rectangle:
(define (draw-rectangle-border x0 y0 width height color)
  (draw-line x0 y0 (+ x0 width -1) y0 color)
  (draw-line x0 y0 x0 (+ y0 height -1) color)
  (draw-line x0 (+ y0 height -1) (+ x0 width -1) (+ y0 height -1) color)
  (draw-line (+ x0 width -1) y0 (+ x0 width -1) (+ y0 height -1) color))

;;; Draw a block with lights and shadows, suitable for Tetris and Breakout;
;;; notice that there is an optional parameter (which is seen by the procedure
;;; as an empty list when not given, and a non-empty list when given):
(define (draw-block x0 y0 width height color . border-size-or-null)
  ;; If the optional parameter was supplied then bind border-size to the
  ;; value given by the user, otherwise use a default:
  (let ((border-size (if (null? border-size-or-null)
                         3
                         (car border-size-or-null))))
    ;; Draw nothing if the block is too small:
    (if (or (<= width (* 2 border-size))
            (<= height (* 2 border-size)))
        #t ; well, we have to return something, so we return #t
        ;; Ok, the block is not too small: draw it:
        (let ((lighter-color (light color)) ; color for the lightened part
              (darker-color (dark color)))  ; color for the shadowed part
          ;; Draw the shadowed part on the East and South sides:
          (draw-rectangle x0 (+ y0 height (- border-size)) width border-size darker-color)
          (draw-rectangle (+ x0 width (- border-size)) y0 border-size height darker-color)
          ;; Draw the enlightended part on the West and Nort sides:
          (draw-rectangle x0 y0 border-size height lighter-color)
          (draw-rectangle x0 y0 width border-size lighter-color)
          ;; Draw the inside part:
          (draw-rectangle (+ x0 border-size) (+ y0 border-size)
                          (- width (* 2 border-size)) (- height (* 2 border-size))
                          color))))) ;; There's no automatic refresh here!

(define (square a)
  (* a a))

;;; The absolute value should not be needed here, but I've found that some rounding
;;; errors with floats make the argument of the square root be negative, and the result
;;; of the square root complex...
;;; If you use this procedure to draw large circles the result will not look good around
;;; the leftmost and rightmost points. But it is reasonably fast, to be used for
;;; interactive applications where speed matters. The result could be made better-looking
;;; by adding another loop on the y variable.
(define (draw-circle center-x center-y radius color)
  (for x (- center-x radius) (+ center-x radius)
       (let ((delta-y (sqrt (abs (- (square radius)
                                    (square (- x center-x)))))))
         (draw-line x (- center-y delta-y) x (+ center-y delta-y) color))))
;;; See the comment above about abs. But notice that this procedure gives good results
;;; also for big circles.
(define (draw-circle-border center-x center-y radius color)
  (for x (- center-x radius) (+ center-x radius)
       (let ((delta-y (sqrt (abs (- (square radius)
                                    (square (- x center-x)))))))
         (draw-pixel x (- center-y delta-y) color)
         (draw-pixel x (+ center-y delta-y) color))))

(define nuance-step -2) ;; Make this bigger in modulo (and still negative) for slow machines
(define (draw-nuanced-circle center-x center-y maximum-radius color)
  (for-step radius maximum-radius 2 nuance-step
    (draw-circle center-x center-y radius
                 (color-* (- 1.2 (/ radius maximum-radius))
                          color))))

(define sqrt2 (sqrt 2))
;;; These can be used as an "approximation" of a circle on slow machines, for interactive
;;; applications which need speed. There is some ugly code duplication here, but the
;;; point of this code is speed, so I don't want to use a higher-order function:
(define (draw-centered-square-border center-x center-y half-diagonal color)
  (let* ((half-side (/ half-diagonal sqrt2))
         (side (* half-side 2))
         (top-left-x (- center-x half-side))
         (top-left-y (- center-y half-side)))
    (draw-rectangle-border top-left-x top-left-y side side color)))
(define (draw-centered-square center-x center-y half-diagonal color)
  (let* ((half-side (/ half-diagonal sqrt2))
         (side (* half-side 2))
         (top-left-x (- center-x half-side))
         (top-left-y (- center-y half-side)))
    (draw-rectangle top-left-x top-left-y side side color)))
