;;; This file is part of guile-with-graphics, a graphic library for Guile

;;; Copyright (C) 2008, 2009  Luca Saiu
;;; Based on turtle, by Luca Saiu

;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.

;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.

;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; This is the most general version of for, but it's not always
;; comfortable to call. I use this to define three friendlier
;; versions:
(define-macro (for-step variable from to step . body)
  (let ((to-name (gensym))
        (from-name (gensym))
        (step-name (gensym))
        (lower-limit-name (gensym))
        (upper-limit-name (gensym)))
    `(let* ((,from-name ,from)
            (,to-name ,to)
            (,variable ,from-name)
            (,lower-limit-name (min ,from-name ,to-name))
            (,upper-limit-name (max ,from-name ,to-name))
            (,step-name ,step))
       (while (and (>= ,variable ,lower-limit-name)
                   (<= ,variable ,upper-limit-name))
              ,@(append body
                        `((set! ,variable (+ ,variable ,step-name))))))))

;; Evaluate the body a number of times equal to how-many-times.
;; No iteration variable is visible in the body,
;;
;; Example: (n-times 5 (display "This will be printed five times\n")
(define-macro (n-times how-many-times . body)
  (let ((variable (gensym)))
    `(for ,variable 1 ,how-many-times ,@body)))

;; The classic for loop, with the variable incremented at each iteration.
;;
;; for example (for i 1 5
;;               (display "i is now ")
;;               (display i)
;;               (newline))
;; is equivalent to the C loop:
;; { 
;;   int i;
;;   for(i = 1; i <= 5; i++){
;;     printf("i is now ");
;;     printf("%i", i);
;;     printf("\n");
;;   } // for
;; }
(define-macro (for variable from to . body)
  `(for-step ,variable ,from ,to 1 ,@body))

;; Just like for, but here the variable is *de*cremented at each iteration:
(define-macro (down-for variable from to . body)
  `(for-step ,variable ,from ,to -1 ,@body))

;;; Return any number in the interval [minimum, maximum], with an uniform
;;; probability distribution:
(define (random-from-to minimum maximum)
  (+ (random (- maximum minimum -1))
     minimum))


;; ;;; Try this:
;; (define (rectangles-demo)
;;   (let ((x0 (random-from-to -10 (+ (window-width) 10)))
;;         (y0 (random-from-to -10 (+ (window-height) 10)))
;;         (width (random-from-to 10 (/ (window-width) 2)))
;;         (height (random-from-to 10 (/ (window-height) 2)))
;;         (color (random-color)))
;;     (handle-quit-event-only)
;;     (lock)
;;     (draw-rectangle x0 y0 width height color)
;;     (unlock)
;;     (refresh-rectangle x0 y0 width height)
;;     (rectangles-demo)))

;; ;;; Try this:
;; (define (rectangle-borders-demo)
;;   (let ((x0 (random-from-to -10 (+ (window-width) 10)))
;;         (y0 (random-from-to -10 (+ (window-height) 10)))
;;         (width (random-from-to 10 (/ (window-width) 2)))
;;         (height (random-from-to 10 (/ (window-height) 2)))
;;         (color (random-color)))
;;     (handle-quit-event-only)
;;     (lock)
;;     (draw-rectangle-border x0 y0 width height color)
;;     (unlock)
;;     (refresh-rectangle x0 y0 width height)
;;     (rectangle-borders-demo)))

;; ;;; Try this:
;; (define (lines-demo)
;;   (let ((x1 (random (window-width)))
;;         (y1 (random (window-height)))
;;         (x2 (random (window-width)))
;;         (y2 (random (window-height)))
;;         (color (random-color)))
;;     (handle-quit-event-only)
;;     (lock)
;;     (draw-line x1 y1 x2 y2 color)
;;     (unlock)
;;     (refresh-rectangle x1 y1 (- x2 x1) (- y2 y1))
;;     (lines-demo)))

;; (define (block-grid-demo)
;;   (let* ((block-width 60)
;;          (block-height 20)
;;          (width (window-width))
;;          (height (window-height)))
;;     (handle-quit-event-only)
;;     (for x 0 (/ width block-width)
;;          (for y 0 (/ height block-height)
;;               (lock)
;;               (draw-block (* x block-width) (* y block-height)
;;                           (- block-width 4) (- block-height 4)
;;                           ;;(dark (dark red))
;;                           (random-color)
;;                           3)
;;               (unlock)))
;;     (refresh)
;;     (block-grid-demo)))

;; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; (define block-width 80)
;; (define block-height 25)
;; (define block-color (dark red))
;; (define block-x (/ (window-width) 2))
;; (define block-y (/ (window-height) 2))
;; (define delta-x 0)
;; (define delta-y 0)
;; (define changing-color #f)
;; (define time-at-startup (time))
;; (define tick-length 30)
;; (define speed (* 0.2 tick-length))

;; (define (handle-event event)
;;   (cond ((or (equal? event '(quit))
;;              (equal? event '(key-down #\q))
;;              (equal? event '(key-down escape)))
;;          (exit 0))
;;         ((equal? event '(key-down left))
;;          (set! delta-x (- speed)))
;;         ((equal? event '(key-up left))
;;          (set! delta-x 0))
;;         ((equal? event '(key-down right))
;;          (set! delta-x speed))
;;         ((equal? event '(key-up right))
;;          (set! delta-x 0))
;;         ((equal? event '(key-down down))
;;          (set! delta-y speed))
;;         ((equal? event '(key-up down))
;;          (set! delta-y 0))
;;         ((equal? event '(key-down up))
;;          (set! delta-y (- speed)))
;;         ((equal? event '(key-up up))
;;          (set! delta-y 0))
;;         ((equal? event '(key-up space))
;;          (set! changing-color #f)) 
;;         ((equal? event '(key-down space))
;;          (set! changing-color #t))
;;         (else
;;          (format #t "Unhandled event: ~a\n" event))))

;; (define (on-tick)
;;   (set! block-x (+ block-x delta-x))
;;   (set! block-y (+ block-y delta-y))
;;   (if changing-color
;;       (set! block-color (random-color)))
;;   (draw-block block-x block-y block-width block-height block-color)
;;   (refresh)
;;   ;;(refresh-rectangle (- block-x 1) (- block-y 1) (+ block-width 2) (+ block-height 2))
;;   (draw-rectangle (- block-x 1) (- block-y 1) (+ block-width 2) (+ block-height 2)
;;                   black))


;; (clear blue)
;; (refresh)

;; ;(rectangles-demo)
;; ;(rectangle-borders-demo)
;; ;(blocks-demo)
;; ;(lines-demo)
;; ;(block-grid-demo)

;; (interactive-game tick-length handle-event on-tick)

;;; Return the n-th element of the given list:
(define (n-th n list)
  (if (zero? n)
      (car list)
      (n-th (- n 1)
            (cdr list))))

;;; Convert the given s-expression into its printer representation:
(define (s-expression->string s-expression)
  (let ((string-port (open-output-string)))
    (write s-expression string-port)
    (get-output-string string-port)))

;;; Some useful constants:
(define pi (* 2 (acos 0)))
(define -pi (- pi))
(define 2pi (* 2 pi))
(define pi/2 (/ pi 2))

;;; Given an angle in radiants, return another equivalent one in (-pi, +pi]
(define (normalize-angle angle)
  (cond ((> angle pi)
         (normalize-angle (- angle 2pi)))
        ((<= angle -pi)
         (normalize-angle (+ angle 2pi)))
        (else
         angle)))

;;; Destructively update a variable representing an angle in radiants,
;;; replacing them with the normalized value:
(define-macro (normalize-angle! variable)
  `(set! ,variable (normalize-angle ,variable)))

;;; Convert an angle from radiants to degrees. This is probably only useful for debugging:
(define (radiants->degrees angle)
  (* 180 (/ angle pi)))

;;; Return a copy of the given list without the given element (if it is in the
;;; original list):
(define (without-with-equality-predicate equality-predicate element list)
  (define (without element list accumulator)
    (cond ((null? list)
           accumulator)
          ((equality-predicate element (car list))
           (append (cdr list) accumulator))
          (else
           (without element (cdr list) (cons (car list) accumulator)))))
  (without element list '()))

;;; The "default" function simply called without uses equal? (structural comparison),
;;; and is just an alias for without-by-structure:
(define (without-by-structure element list)
  (without-with-equality-predicate equal? element list))
(define without without-by-structure)
 
;;; without-by-identity uses eq? (equality by identity)
(define (without-by-identity element list)
  (without-with-equality-predicate eq? element list))

;;; Return a copy of the first parameter without the elements in the second parameter;
;;; the equality is by structure. This is an implementation of set subtraction, where
;;; sets are implemented as unsorted lists.
(define (subtract-lists list elements-to-remove)
  (if (null? elements-to-remove)
      list
      (subtract-lists (without-by-structure (car elements-to-remove) list)
                      (cdr elements-to-remove))))

;;; Call the given function on the elements of the given list, in order. The results
;;; are ignored:
(define (iterate function list)
  (if (null? list)
      #t
      (begin
        (function (car list))
        (iterate function (cdr list)))))

;;; Common Lisp-style when and unless. These are one-way if conditionals, which take
;;; one or more forms as their body even without using begin. when executes its body
;;; (and returns the value of its last form) if the condition is true, and unless
;;; executes its body (and returns the value of its last form) if the condition is
;;; false:
(define-macro (when condition . body)
  `(if ,condition
       (begin
         ,@body)))
(define-macro (unless condition . body)
  `(if (not ,condition)
       (begin
         ,@body)))

;;; Useful higher-order functions working on lists (note that map and filter
;;; are predefined):
(define (for-all? predicate? list)
  (cond ((null? list)
         #t)
        ((predicate? (car list))
         (for-all? predicate? (cdr list)))
        (else
         #f)))
(define (exists? predicate? list)
  (cond ((null? list)
         #f)
        ((predicate? (car list))
         #t)
        (else
         (exists? predicate? (cdr list)))))

;;; Executes the forms in body and return the result of the last one. If any
;;; of them throws an exception, just capure it, stop executing the body and
;;; return #f:
(define-macro (ignore-failures . body)
  `(catch #t
     (lambda ()
       ,@body)
     (lambda who-cares
       #f)))

(define (square x)
  (* x x))

;;; Compute the Euclidean distance between two points:
(define (distance x1 y1 x2 y2)
  (sqrt (+ (square (- x1 x2))
           (square (- y1 y2)))))

;;; Return x if it's within the given bounds (included), otherwise return the
;;; limit:
(define (saturate x minimum maximum)
  (cond ((< x minimum)
         minimum)
        ((> x maximum)
         maximum)
        (else
         x)))

;;; Set the initial random state to a value depending on the current time,
;;; so that every time we start up the program the sequence of random values
;;; is different:
(define (randomize-random-state!)
  (let ((time (gettimeofday)))
    (set! *random-state*
          (seed->random-state (+ (car time)
                                 (cdr time))))))

;;; This is very useful for debugging. The given expression is evaluated and printed
;;; on the output in a clearly readable format. For example (log (+ 1 2)) prints
;;; "(+ 1 2): 3" on the output.
(define-macro (log expression)
  `(format #t "~s: ~s\n" ',expression ,expression))

;;; The given procedure is called in an infinite loop, at most
;;; maximum-frequency times per second (less frequently on slower
;;; machines). The given procedure is given the time elapsed since
;;; the last call in seconds as its only parameter.
(define (call-with-maximum-frequency maximum-frequency procedure)
  (let* ((minimum-period (/ 1000 maximum-frequency)) ; in milliseconds
         (minimum-period-in-seconds (/ minimum-period 1000))
         (time-before-last-period (milliseconds-since-startup)))
    (while #t
      (let* ((time-before-this-period (milliseconds-since-startup))
             (delta-t (- time-before-this-period
                         time-before-last-period))
             (milliseconds-to-sleep (- minimum-period delta-t)))
        (if (> milliseconds-to-sleep 0)
            (begin
              (usleep (inexact->exact (floor (* milliseconds-to-sleep 1000))))
              (set! time-before-last-period (+ time-before-this-period
                                               milliseconds-to-sleep))
              (procedure minimum-period-in-seconds))
            (begin
              (usleep 0) ; this should at least yield to another process
              (set! time-before-last-period time-before-this-period)
              (procedure (/ delta-t 1000))))))))
